import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/logger/logger.dart';
import 'package:contentsquare/src/modules/screengraph/compression.dart';
import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
import 'package:fast_immutable_collections/fast_immutable_collections.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import '../screengraph/hit_test/get_element_data_of_hit_render_object.dart';
import '../screengraph/hit_test/hit_element_data.dart';
import 'exceptions/exceptions.dart';
import 'logs/logs.dart';
import 'screengraph_retrieval.dart';

/// Widget used to track a WebView.
///
/// {@template contentsquare.webview.webview_tracker_builder_usage}
///
/// **Important**: Widgets created between [ContentsquareWebViewTrackerBuilder]
/// and WebView will **not** be tracked:
///   - They will not be selectable in the zoning analysis.
///   - Gestures on those widgets will not be tracked.
/// Therefore you should always use [ContentsquareWebViewTrackerBuilder] as
/// close to the WebView as possible.
/// {@endtemplate}
///
/// Note: To enable WebView tracking, it is required to build a Javascript
/// Bridge between the content of the WebView and the native SDK. To do so,
/// you will have to implement the Contentsquare WebView Javascript Tracking
/// Tag in the web pages called in your app WebViews.
///   - See https://docs.contentsquare.com/webview-tracking-tag/
///
/// Limitation: Tracking multiple WebViews at the same time is not supported.
///
/// {@endtemplate}
class ContentsquareWebViewTrackerBuilder extends StatefulWidget {
  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  ///
  /// Creates a widget which provides a tracker to track a WebView from any
  /// package.
  ///
  /// For the tracking to work, you need to:
  ///   - Enable javascript in the WebView.
  ///   - Register the given [ContentsquareWebViewTracker.javascriptChannels]
  ///     to allow the webview to communicate with the SDK.
  ///   - Use [ContentsquareWebViewTracker.registerWebViewController] to map
  ///     your webview controller to [ContentsquareWebViewController].
  ///     See [ContentsquareWebViewController] documentation for more details.
  ///   - Use [ContentsquareWebViewTracker.startPageTracking] to start
  ///     the tracking.
  ///     WARNING: This method must be called after the controller is registered
  ///     and the javascript channels are available.
  ///
  /// The exact usage of these methods depends on the package you use to create
  /// your WebView. Reach out to us if you need help.
  ///
  /// {@endtemplate}
  const ContentsquareWebViewTrackerBuilder({
    Key? key,
    required this.builder,
  }) : super(key: key);

  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  @Deprecated('Use ContentsquareWebViewTrackerBuilder instead')
  const factory ContentsquareWebViewTrackerBuilder.custom({
    Key? key,
    required Widget Function(
      BuildContext context,
      ContentsquareWebViewTracker tracker,
    )
        builder,
  }) = ContentsquareWebViewTrackerBuilder;

  /// Callback in which the WebView should be built.
  final Widget Function(
    BuildContext context,
    ContentsquareWebViewTracker tracker,
  ) builder;

  @override
  State<ContentsquareWebViewTrackerBuilder> createState() =>
      _a166();
}


class _a166
    extends State<ContentsquareWebViewTrackerBuilder> {
  
  late final a37 = a169(
    a34: a432,
    a33: _a274,
  );

  
  String? _a274() {
    
    
    
    
    if (!mounted) {
      a212.a30(a158());
      return null;
    }

    
    
    
    
    
    
    final a433 = context as Element;

    
    
    
    
    
    
    
    
    
    
    final a434 = a433.renderObject;
    if (a434 == null) {
      a212.a30(a157());
      return null;
    }
    if (a434 is! a168) {
      a212.a26(
        'The type of render object associated with the $ContentsquareWebViewTrackerBuilder widget is ${a434.runtimeType}.\n'
        'This is an error, it should always be a $a168.\n',
      );
      return null;
    }

    
    
    
    
    
    
    
    
    final a435 = a434.a262;
    if (a435 == null) {
      a212.a30(a156());
      return null;
    }

    final a436 = a231(
      a536: a434,
      a537: a126.a242(
        a546: a435.center,
      ),
    );

    return a436.a12(
      a467: (hitElementData) {
        return a219(hitElementData.a65);
      },
      
      
      a466: () => null,
    );
  }

  @override
  void dispose() {
    a37._a281();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _a170(
      a582: _a167(
        a577: widget.builder(context, a37),
        a36: a37.a8,
      ),
    );
  }
}


class _a167
    extends SingleChildRenderObjectWidget {
  const _a167({
    Key? a576,
    required this.a36,
    required Widget a577,
  }) : super(key: a576, child: a577);

  
  final Future<Map<String, dynamic>> Function() a36;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return a168(a35: a36);
  }
}


class a168 extends RenderProxyBox {
  
  a168({
    required this.a35,
  });

  
  final Future<Map<String, dynamic>> Function() a35;
}


class a169 implements ContentsquareWebViewTracker {
  
  a169({
    required this.a34,
    required this.a33,
  });

  
  final a165 a34;

  
  final String? Function() a33;

  
  ContentsquareWebViewController? _a437;

  @override
  void registerWebViewController(ContentsquareWebViewController controller) {
    _a437 = controller;
  }

  @override
  Future<void> startPageTracking() async {
    final _a438 = this._a437;

    if (_a438 == null) {
      return a212.a30(
        a151(
          a568: null,
          a569: 'WebViewController is null, did you forget to use '
              'tracker.onWebViewCreated?',
        ),
      );
    }

    
    final a439 = _a438.isChannelAvailable;
    if (a439 != null) {
      final bool a440;
      try {
        a440 =
            await _a276(
          a579: a439,
        );
      } catch (error) {
        return a212.a30(
          a151(
            a568: await _a438.currentUrl?.call(),
            a569: 'Unexpected error while executing JS code, check the '
                'registration of the javascript channels: $error',
          ),
        );
      }
      if (!a440) {
        return a212.a30(
          a151(
            a568: await _a438.currentUrl?.call(),
            a569:
                'JavascriptChannels are not registered, did you forget to use '
                'tracker.javascriptChannels?',
          ),
        );
      }
    }

    try {
      await _a438.evaluateJavascript(
        '''
${_a275(a578: _a438.javascriptChannelCall)}
        
// This is needed because [_webViewController.evaluateJavascript] only works on
// js code that returns a value (and the bridge registration does not return a 
// value)
true;
''',
      );
    } catch (error) {
      return a212.a30(
        a151(
          a568: await _a438.currentUrl?.call(),
          a569: 'Unexpected error while executing JS code to setup the '
              'Javascript bridge: $error',
        ),
      );
    }

    return a212.a30(
      a152(
        a570: await _a438.currentUrl?.call(),
      ),
    );
  }

  @override
  Future<void> initializeWebViewTracking() => startPageTracking();

  @override
  Set<ContentsquareJavascriptChannel> get javascriptChannels {
    return ContentsquareJavascriptEventEnum.values
        .map(
          (event) => a150(
            a46: Contentsquare(),
            a45: event,
            a44: a33,
          ).a2,
        )
        .toSet();
  }

  
  String _a275({
    required String Function(String channelName, String data)
        a578,
  }) {
    return '''
window.CSJavascriptBridge = {
  
  ${ContentsquareJavascriptEventEnum.values.map(
              (event) => '${event.name}: ${a150(
                a46: Contentsquare(),
                a45: event,
                a44: a33,
              ).a9(
                a454: a578,
              )}',
            ).join(',\n\n').replaceAll('\n', '\n  ')}
};''';
  }

  
  Future<bool> _a276({
    required Future<bool> Function(String) a579,
  }) async {
    for (final a441 in javascriptChannels) {
      final a442 =
          await a579(a441.name);

      if (!a442) {
        return false;
      }
    }

    return true;
  }

  
  
  
  
  var _a443 = IList<Stopwatch>();

  
  Stopwatch _a277() {
    final a444 = Stopwatch();
    _a443 = _a443.add(a444);
    return a444;
  }

  
  void _a278(Stopwatch a580) {
    a580.stop();
    _a443 = _a443.remove(a580);
  }

  
  @visibleForTesting
  static String get a0 => '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      window.cs_wvt.push(['serializeWebView']);
    }
  ''';

  
  Future<Map<String, dynamic>> a8() async {
    final a445 = _a437;
    if (a445 == null) {
      throw a149();
    }

    
    
    final a446 = _a277();
    a446.start();

    bool a279() {
      return a446.elapsed >=
          a34.a39;
    }

    
    
    
    
    bool a280() {
      return !a446.isRunning;
    }

    Map<String, dynamic>? a447;
    
    
    
    
    while (!(a279() || a280())) {
      final a448 = await a445.evaluateJavascript(
        a0,
      );
      final Object? a449;
      try {
        final a450 = jsonDecode(a448);
        if (a450 is String) {
          
          
          
          a449 = jsonDecode(a450);
        } else {
          a449 = a450;
        }
      } on FormatException catch (e) {
        a212.a30(
          a161(
            a43: a448,
            a42: e,
          ),
        );
        continue;
      }

      
      
      
      
      
      
      
      final a451 = a449 == null ||
          a449 is num ||
          a449 is Map<String, dynamic>;
      if (!a451) {
        a212.a30(
          a162(
            a41: a449,
          ),
        );
      }

      if (a449 is Map<String, dynamic>) {
        a447 = a449;
        break;
      }

      await Future.delayed(
        a34.a38,
      );
    }

    _a278(a446);

    if (a447 == null) {
      throw a147(
        a47: a34.a39,
      );
    }

    return a447;
  }

  
  void _a281() {
    _a443.forEach(_a278);
  }
}


class _a170 extends InheritedWidget {
  const _a170({
    Key? a581,
    required Widget a582,
  }) : super(key: a581, child: a582);

  @override
  bool updateShouldNotify(oldWidget) => false;
}


extension a179 on Element {
  
  bool get _a282 {
    return widget is ContentsquareWebViewTrackerBuilder;
  }

  
  bool get a283 {
    final a452 =
        getElementForInheritedWidgetOfExactType<_a170>();
    return a452 != null || _a282;
  }
}
